unit WaveOut;

//------------------------------------------------------------------------------
//
//  A simple component for wave out with LibAO under Linux
//  Done by Paul Limanowka (contact: p.limanowka@borland.pl)
//
//------------------------------------------------------------------------------

interface

uses
  SysUtils, Types, Classes, ao;

type
  TWaveEvent = procedure(Sender:tObject; Buf:pointer; BufLen:integer) of object;

  TWaveThread = class(tThread)
  private
    fBuffers : pointer;
    fBufDuration : integer;
    fBufLen : integer;
    fBufCnt : integer;
    fPlayingBuf : integer;
    fWriteBuf : integer;
    fDev : pAoDevice;
    fOnBuf : TWaveEvent;
    procedure DoWaveBuf;
  public
    constructor Create(dev:string; rate,bits,channels,BufLen,BufCount:integer);
    procedure Execute; override;
    destructor Destroy; override;
    property OnDataNeeded:TWaveEvent read fOnBuf write fOnBuf;
  end;

  TWaveOut = class(TComponent)
  private
    { Private declarations }
    fPlayThread: tWaveThread;
    fDevs: tStringList;
    fOnBuf: tWaveEvent;
    fBits: integer;
    fRate: integer;
    fChn: integer;
    fDrv: string;
    fBufLen: integer;
    fBufCnt: integer;
    procedure PlayThreadTerminated(Sender:tObject);
    procedure SetOnBuf(const Value: TWaveEvent);
    procedure SetBits(const Value: integer);
    procedure SetChn(const Value: integer);
    procedure SetDrv(const Value: string);
    procedure SetRate(const Value: integer);
    procedure SetBufCnt(const Value: integer);
    procedure SetBufLen(const Value: integer);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(aOwner:tComponent); override;
    procedure Play;
    procedure Pause;
    procedure Stop;
    destructor Destroy; override;
    property Devices:tStringList read fDevs;
  published
    { Published declarations }
    property OnBufNeeded: TWaveEvent read fOnBuf write SetOnBuf default nil;
    property Device:string read fDrv write SetDrv;
    property Frequency:integer read fRate write SetRate default 44100;
    property Bits:integer read fBits write SetBits default 16;
    property Channels:integer read fChn write SetChn default 2;
    property BufferLen:integer read fBufLen write SetBufLen default 8820;
    property BuffersCount:integer read fBufCnt write SetBufCnt default 8;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Additional', [TWaveOut]);
end;

{ tWaveThread }

constructor tWaveThread.Create(dev: string; rate, bits, channels, BufLen,
  BufCount: integer);
var drvId:integer;
begin
  inherited Create(true);
  FreeOnTerminate:=true;
  fOnBuf:=nil;
  fBufCnt:=BufCount;
  fBufLen:=BufLen;
  fBuffers:=AllocMem(fBufLen*fBufCnt);
  fPlayingBuf:=0;
  fWriteBuf:=0;

  fBufDuration:=(rate*channels*(bits div 8)*1000) div fBufLen;

  drvId:=aoGetDriverId(dev);
  if drvId<>-1 then
    fDev:=aoOpen(drvId,bits,rate,channels,nil)
  else
    fDev:=nil;
end;

destructor tWaveThread.Destroy;
begin
  if Assigned(fDev) then aoClose(fDev);
  inherited;
end;

procedure tWaveThread.DoWaveBuf;
begin
  if Assigned(fOnBuf) then
    fOnBuf(Self,pointer(integer(fBuffers)+(fWriteBuf*fBufLen)),fBufLen);
  fWriteBuf:=(fWriteBuf+1) mod fBufCnt;
end;

procedure tWaveThread.Execute;
begin
  repeat
    if Assigned(fDev) then
      aoPlay(fDev,pointer(integer(fBuffers)+(fPlayingBuf*fBufLen))^,fBufLen)
    else
      sleep(fBufDuration);
    fPlayingBuf:=(fPlayingBuf+1) mod fBufCnt;
    Synchronize(DoWaveBuf);
  until terminated;
end;

{ TWaveOut }

constructor TWaveOut.Create(aOwner: tComponent);
var i:integer;
begin
  inherited Create(aOwner);
  fDevs:=tStringList.Create;
  fPlayThread:=nil;
  fOnBuf:=nil;

  for i:=low(drvNames) to high(drvNames) do
    if aoGetDriverId(drvNames[i])<>-1 then
      fDevs.Add(drvNames[i]);

  fDrv:=''; fRate:=44100; fBits:=16; fChn:=2; fBufLen:=8820; fBufCnt:=8;
end;

destructor TWaveOut.Destroy;
begin
  if Assigned(fPlayThread) then fPlayThread.Terminate;
  fDevs.Destroy;
  inherited;
end;

procedure TWaveOut.Pause;
begin
  if Assigned(fPlayThread) and not fPlayThread.Suspended then
    fPlayThread.Suspend;
end;

procedure TWaveOut.Play;
begin
  if not Assigned(fPlayThread) then
  begin
    fPlayThread:=tWaveThread.Create(fDrv,fRate,fBits,fChn,fBufLen,fBufCnt);
    fPlayThread.OnTerminate:=PlayThreadTerminated;
    fPlayThread.OnDataNeeded:=fOnBuf;
  end;
  if fPlayThread.Suspended then
    fPlayThread.Resume;
end;

procedure TWaveOut.Stop;
begin
  if Assigned(fPlayThread) then
    fPlayThread.Terminate;
end;

procedure TWaveOut.PlayThreadTerminated(Sender: tObject);
begin
  fPlayThread:=nil;
end;

procedure TWaveOut.SetOnBuf(const Value: tWaveEvent);
begin
  fOnBuf := Value;
  if Assigned(fPlayThread) then fPlayThread.OnDataNeeded:=fOnBuf;
end;

procedure TWaveOut.SetBits(const Value: integer);
begin
  if not Assigned(fPlayThread) then
    fBits := Value;
end;

procedure TWaveOut.SetChn(const Value: integer);
begin
  if not Assigned(fPlayThread) then
    fChn := Value;
end;

procedure TWaveOut.SetDrv(const Value: string);
begin
  if not Assigned(fPlayThread) then
    fDrv := Value;
end;

procedure TWaveOut.SetRate(const Value: integer);
begin
  if not Assigned(fPlayThread) then
    fRate := Value;
end;

procedure TWaveOut.SetBufCnt(const Value: integer);
begin
  if not Assigned(fPlayThread) then
    fBufCnt := Value;
end;

procedure TWaveOut.SetBufLen(const Value: integer);
begin
  if not Assigned(fPlayThread) then
    fBufLen := Value;
end;

initialization
  ao.aoInitialize;
finalization
  ao.aoShutdown;
end.
